// -*- C++ -*-
// $RCSfile: bicg.h,v $
// $Revision: 1.16 $
// $Author: vrc $
// $Date: 2007-09-26 22:03:10 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include "common/tostring.h"
#include "common/IO/progressbar.h" 
#include "engine/log_value.h"
//*****************************************************************
// Iterative template routine -- BiCG
//
// BiCG solves the unsymmetric linear system Ax = b 
// using the Preconditioned BiConjugate Gradient method
//
// BiCG follows the algorithm described on p. 22 of the 
// SIAM Templates book.
//
// The return value indicates convergence within max_iter (input)
// iterations (0), or no convergence within max_iter iterations (1).
//
// Upon successful return, output arguments have the following values:
//  
//        x  --  approximate solution to Ax = b
// max_iter  --  the number of iterations performed before the
//               tolerance was reached
//      tol  --  the residual after the final iteration
//  
//*****************************************************************


template < class Matrix, class Vector, class Preconditioner, class Real >
int 
BiCG(const Matrix &A, Vector &x, const Vector &b,
     const Preconditioner &M, int &max_iter, Real &tol)
{
  //Thread control starts here
  Progress * MyController = getProgress();
  std::string out_mssg;
  //Thread control ends here
  Real resid;
  Vector rho_1(1), rho_2(1), alpha(1), beta(1);
  Vector z, ztilde, p, ptilde, q, qtilde;

  Real normb = norm(b);
  Vector r = b - A * x;
  Vector rtilde = r;

  if (normb == 0.0)
    normb = 1;
  
  if ((resid = norm(r) / normb) <= tol) {
    //Thread control starts here
    out_mssg = "solving: " + to_string(resid) + "/" + to_string(tol);
    MyController->set_message(out_mssg);
    //MyController->set_success(); // notify that BiCG finished successfully (woo-hoo!)
    //MyController.finalize_threads();
    // tell the parent thread to stop monitoring and move on to next task
    //Thread control ends here
    tol = resid;
    max_iter = 0;
    return 0;
  }
  LogValue doubletolog(resid, tol);
  for (int i = 1; i <= max_iter && !MyController->query_stop(); i++) {
    z = M.solve(r);
    ztilde = M.trans_solve(rtilde);
    rho_1(0) = dot(z, rtilde);
    if (rho_1(0) == 0) { 
      tol = norm(r) / normb;
      max_iter = i;
      //Thread control starts here
      MyController->set_failure(); // notify that BiCG finished successfully (woo-hoo!)
      //Thread control ends here
      return 2; 
    }
    if (i == 1) {
      p = z;
      ptilde = ztilde;
    } else {
      beta(0) = rho_1(0) / rho_2(0);
      p = z + beta(0) * p;
      ptilde = ztilde + beta(0) * ptilde;
    }
    q = A * p;
    qtilde = A.trans_mult(ptilde);
    alpha(0) = rho_1(0) / dot(ptilde, q);
    x += alpha(0) * p;
    r -= alpha(0) * q;
    rtilde -= alpha(0) * qtilde;

    rho_2(0) = rho_1(0);
    if ((resid = norm(r) / normb) < tol) {
      //Thread control starts here
      MyController->set_message(to_string(resid) + "/" + to_string(tol));
      //MyController->set_success(); // notify that BiCG finished successfully (woo-hoo!)
      //Thread control ends here
      tol = resid;
      max_iter = i;
      return 0;
    }
    //Thread control starts here
    out_mssg = "solving: " + to_string(resid) + "/" + to_string(tol);
    MyController->set_message(out_mssg);
    MyController->set_progress(doubletolog(resid)); /* 
						      Thread controller returns the current relative
						      residual, which will be used to construct a 
						      graphic or text bar.
						   */
    //Thread control ends here
  }

  tol = resid;
  //Thread control starts here
  MyController->set_failure(); //If thread made it to this point, it means that solver failed.
  //Thread control ends here
  return 1;
}
  
